import requests
import urllib.parse
from omnik.ha_logger import hybridlogger
import hashlib
import xml.etree.ElementTree as ET
import datetime
from omnik.plugin_client import Client


class SolarmanPVClient(Client):
    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "Client enabled: solarmanpvclient"
        )

        # API Key
        self.api_key = self.config.get(
            "client.solarmanpv", "api_key", fallback="apitest"
        )
        self.user_id = -1
        self.token = None

        self.base_url = self.config.get(
            "client.solarmanpv",
            "base_url",
            fallback="http://www.solarmanpv.com:18000/SolarmanApi/serverapi",
        )

        self.username = self.config.get("client.solarmanpv", "username")
        self.password = self.config.get("client.solarmanpv", "password")
        # Get plant_id_list
        self.plant_id_list = self.config.getlist(
            "client.solarmanpv", "plant_id_list", fallback=[]
        )

    def initialize(self):
        pwhash = hashlib.md5(self.password.encode("utf-8")).hexdigest()
        url = f"{self.base_url}/Login?username={self.username}&password={pwhash}&key={self.api_key}"

        xmldata = self._api_request(url, "GET", None)
        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", f"account validation: {xmldata}"
        )

        # Parsing the XML
        for element in ET.fromstring(xmldata):
            if element.tag == "userID":
                self.user_id = int(element.text)
            elif element.tag == "token":
                self.token = element.text

    def _api_request(self, url, method, body, encode=False):
        """
        This API mithod uses XML. When logging in a token is obtained.
        For each request is to be checked if the token is still valid.
        """
        headers = {"Content-type": "application/x-www-form-urlencoded"}

        data = urllib.parse.urlencode(body) if encode and body is not None else body

        r = requests.request(method=method, url=url, data=data, headers=headers)

        r.raise_for_status()

        return r.content

    def getPlants(self):

        data = []
        for plant in self.plant_id_list:
            data.append({"plant_id": plant})

        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", f"plant list from config {data}"
        )

        return data

    def xmlprop(self, xml, params, fallback=None):
        paramlist = params
        param = paramlist.pop(0)
        for el in xml:
            if param == el.tag:
                if paramlist:
                    return self.xmlprop(el, paramlist, fallback)
                else:
                    return el.text
        return fallback

    def getPlantData(self, plant_id):

        data = {}
        url = f"{self.base_url}/Data?username={self.username}&stationid={plant_id}&key={self.api_key}&token={self.token}"

        xmldata = self._api_request(url, "GET", None)
        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", f"plant data ({plant_id}) {xmldata}"
        )
        # Parsing the XML
        xml = ET.fromstring(xmldata)

        # Check if an error was returned
        if xml.tag != "data":
            # log warning
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                "No valid data received. Trying to renew token.",
            )
            # Get new token
            self.initialize()
            # Retry the call
            url = f"{self.base_url}/Data?username={self.username}&stationid={plant_id}&key={self.api_key}&token={self.token}"
            xmldata = self._api_request(url, "GET", None)
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                f"plant data retry ({plant_id}) {xmldata}",
            )
            # Parsing the XML
            xml = ET.fromstring(xmldata)

        if xml.tag == "data":
            data["name"] = self.xmlprop(xml, ["name"])
            data["income"] = float(self.xmlprop(xml, ["income", "TotalIncome"], 0.0))
            data["data_logger"] = self.xmlprop(xml, ["detail", "WiFi", "id"], "")
            data["inverter"] = self.xmlprop(
                xml, ["detail", "WiFi", "inverter", "SN"], ""
            )
            data["status"] = int(
                self.xmlprop(xml, ["detail", "WiFi", "inverter", "status"], 0)
            )
            data["current_power"] = int(
                float(self.xmlprop(xml, ["detail", "WiFi", "inverter", "power"], 0.0))
                * 1000
            )
            data["today_energy"] = float(
                self.xmlprop(xml, ["detail", "WiFi", "inverter", "etoday"], 0.0)
            )
            data["total_energy"] = float(
                self.xmlprop(xml, ["detail", "WiFi", "inverter", "etotal"], 0.0)
            )
            data["last_update"] = float(
                self.xmlprop(
                    xml,
                    ["detail", "WiFi", "inverter", "lastupdated"],
                    datetime.datetime.now().timestamp(),
                )
            )
            data["last_update_time"] = datetime.datetime.utcfromtimestamp(
                data["last_update"]
            ).strftime("%Y-%m-%dT%H:%M:%SZ")
            return data
        else:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "ERROR", f"Cannot fetch data: {xmldata}"
            )
            return None



from omnik.ha_logger import hybridlogger
import omnik.InverterMsg
from omnik.plugin_client import Client
import socket
import binascii


class TCPclient(Client):

    # IMPORTANT : ONLY WIFI Modules with s/n 602xxxxxx and 604xxxxxx support direct access thru port 8899!
    # Based on: https://github.com/woutrrr/omnik-data-logger by Wouter van der Zwan
    # The script was tested only using simulations, let me know if this works for you

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "Client enabled: TCPclient"
        )

        # Get plant_id_list
        self.plant_id_list = self.config.getlist(
            "client.tcpclient", "plant_id_list", fallback=[]
        )
        if not self.plant_id_list:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "plant_id_list was not specified for client TCPclient",
            )
            raise Exception("plant_id_list was not specified")
        # Initialize dict with inverter info
        self.inverters = {}

    def getPlants(self):
        data = []
        inverterdata = {}
        for plant in self.plant_id_list:
            inverter_address = self.config.get(
                f"plant.{plant}", "inverter_address", fallback=None
            )
            if not inverter_address:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    f"inverter_address for plant {plant} was not specified [TCPclient]",
                )
                raise Exception("an inverter_address was not specified")
            inverter_port = int(
                self.config.get(f"plant.{plant}", "inverter_port", fallback="8899")
            )
            inverter_connection = (inverter_address, int(inverter_port))
            logger_sn = self.config.get(f"plant.{plant}", "logger_sn", fallback=None)
            if not logger_sn:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    "logger_sn (The serial number of the "
                    "Wi-Fi datalogger) for plant {plant} was not specified for [TCPclient]",
                )
                raise Exception(
                    "logger_sn (a serial number f the Wi-Fi datalogger) was not specified"
                )
            inverter_sn = self.config.get(
                f"plant.{plant}", "inverter_sn", fallback=None
            )
            if not inverter_sn:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    "inverter_sn (The serial number of the inverter) "
                    "for plant {plant} was not specified for [TCPclient]",
                )
                raise Exception(
                    "inverter_sn (a serial number of the inverter) was not specified"
                )
            inverterdata = {
                "inverter_address": inverter_address,
                "inverter_port": inverter_port,
                "logger_sn": int(logger_sn),
                "inverter_sn": inverter_sn,
                "inverter_connection": inverter_connection,
            }
            self.inverters[plant] = inverterdata
            data.append({"plant_id": plant})

        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", f"plant list from config {data}"
        )

        return data

    def getPlantData(self, plant_id):

        valid = False
        data = {}
        # Create request message
        requestmsg = omnik.InverterMsg.request_string(
            self.inverters[plant_id]["logger_sn"]
        )
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(40)
        try:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                "Connecting to %s port %s"
                % self.inverters[plant_id]["inverter_connection"],
            )
            sock.connect(self.inverters[plant_id]["inverter_connection"])
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                "Sending request message with logger_sn: "
                f"{omnik.InverterMsg.request_string(self.inverters[plant_id]['logger_sn'])}",
            )
            # Sending request message
            sock.sendall(requestmsg)
            # Wait for response (listen and block thread)
            rawmsg = sock.recv(129)
            # Validate the data response. Check length and serial number
            if len(rawmsg) >= 99:
                inverterMsg = omnik.InverterMsg.InverterMsg(rawmsg)
                serialnr = inverterMsg.getID()
                # lookup plant_id to see it is in the plants list
                if self.inverters[plant_id]["inverter_sn"] == serialnr:
                    data["plant_id"] = plant_id
                    valid = True
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    f"data size: {len(rawmsg)}.\n"
                    f"Datadump: {str(binascii.b2a_base64(rawmsg))}",
                )
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Inverter is not reachable, this is normal when it is dark. {e}",
            )
            return None

        if valid:
            # Get the data from the received message
            inverterMsg.FetchDataDict(data)
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"New message received from inverter '{serialnr}'",
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                "Invalid response, ignoring message",
            )
            data = None
        return data



import requests
from omnik.ha_logger import hybridlogger

from omnik.plugin_output import Plugin
import threading


class influxdb(Plugin):
    def __init__(self):
        super().__init__()
        self.name = "influxdb"
        self.description = "Write output to InfluxDB"
        self.host = self.config.get("output.influxdb", "host", fallback="localhost")
        self.port = self.config.get("output.influxdb", "port", fallback="8086")
        self.database = self.config.get(
            "output.influxdb", "database", fallback="omnikdatalogger"
        )

        self.headers = {"Content-type": "application/x-www-form-urlencoded"}
        # Add JWT authentication token
        if self.config.has_option("output.influxdb", "jwt_token"):
            self.auth = None
            self.headers[
                "Authorization"
            ] = f"Bearer {self.config.get('output.influxdb', 'jwt_token')}"
        elif self.config.has_option(
            "output.influxdb", "username"
        ) and self.config.has_option("output.influxdb", "password"):
            self.auth = requests.auth.HTTPBasicAuth(
                self.config.get("output.influxdb", "username"),
                self.config.get("output.influxdb", "password"),
            )
        else:
            self.auth = None
        self.timestamp_field = {}
        for field in self.config.data_field_config:
            if not self.config.data_field_config[field]["dev_cla"] == "timestamp":
                continue
            self.timestamp_field[self.config.data_field_config[field]["asset"]] = field
        self.access = threading.Condition(threading.Lock())

    def _get_temperature(self, values):
        if self.config.getboolean("output.influxdb", "use_temperature", fallback=False):
            weather = self.get_weather()
            values["temperature"] = str(weather["main"]["temp"])
            values["timestamp_ow"] = weather["main"]["dt"]

    def _get_attributes(self, values, asset_class):
        attributes = {}
        for att in self.config.attributes["asset"][asset_class]:
            if att in self.config.data_field_config:
                if self.config.data_field_config[att]["dev_cla"] == "timestamp":
                    continue
            attributes[att] = values[att]
        return attributes.copy()

    def _get_tags(self, field, attributes, asset_class):
        tags = attributes.copy()
        tags["asset"] = asset_class
        tags["entity"] = field
        tags["name"] = str(self.config.data_field_config[field]["name"]).replace(
            " ", "\\ "
        )
        if self.config.data_field_config[field]["dev_cla"]:
            tags["dev_cla"] = self.config.data_field_config[field]["dev_cla"]
        if self.config.data_field_config[field]["unit"]:
            tags["unit"] = str(self.config.data_field_config[field]["unit"]).replace(
                " ", "\\ "
            )
        tags.update(self.config.data_field_config[field]["tags"])
        return tags.copy()

    def _format_output(self, field, values):
        if field not in self.config.data_field_config:
            return ""
        if not self.config.data_field_config[field]["measurement"]:
            # no measurement, exclude
            return ""

        if self.config.data_field_config[field]["dev_cla"] == "timestamp":
            # skip the epoch fields
            return ""

        asset_class = self.config.data_field_config[field]["asset"]
        attributes = self._get_attributes(values, asset_class)
        nanoepoch = int(values[self.timestamp_field[asset_class]] * 1000000000)

        if field in values and self.config.data_field_config:
            # Get tags
            tags = self._get_tags(field, attributes, asset_class)
            # format output data using the InfluxDB line protocol
            # https://v2.docs.influxdata.com/v2.0/write-data/#line-protocol
            return (
                f'{self.config.data_field_config[field]["measurement"]},'
                f'{",".join("{}={}".format(key, value) for key, value in tags.items())} '
                f"value={values[field]} {nanoepoch}\n"
            )
        else:
            return ""

    def process(self, **args):
        # Send data to influxdb
        try:
            self.access.acquire()
            msg = args["msg"]
            hybridlogger.ha_log(
                self.logger, self.hass_api, "DEBUG", f"Logging data to InfluxDB: {msg}"
            )
            values = msg.copy()
            # self._get_temperature(values)

            # Build structure
            encoded = ""
            for field in values:
                encoded += self._format_output(field, values)

            # Influx has no tables! Use measurement prefix
            # encoded = f'inverter,plant=p1 {",".join("{}={}".format(key, value)
            # for key, value in values.items())} {nanoepoch}'

            # curl -i -XPOST 'http://localhost:8086/write?db=mydb' --data-binary
            # 'cpu_load_short,host=server01,region=us-west value=0.64 1434055562000000000'
            url = f"http://{self.host}:{self.port}/write?db={self.database}"

            r = requests.post(url, data=encoded, headers=self.headers, auth=self.auth)

            r.raise_for_status()
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                f"Submit to Influxdb {url} successful.",
            )
        except requests.exceptions.HTTPError as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Got error from influxdb: {e.args} (ignoring: if this happens a lot ... fix it)",
            )
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Got unknown submitting to influxdb: {e.args} (ignoring: if this happens a lot ... fix it)",
            )
        finally:
            self.access.release()



from datetime import datetime, tzinfo
import pytz
import json
import time
from omnik.ha_logger import hybridlogger

import paho.mqtt.client as mqttclient
import uuid

from omnik.plugin_output import Plugin
import threading

MEASUREMENT_CLASSES = [
    "power",
    "current",
    "temperature",
    "voltage",
    "frequency",
    "flow",
]
MEASUREMENT_TOTAL_INCREASING_CLASSES = ["energy", "gas", "operation_time", "count"]
STATE_CLASS_MEASUREMENT = "measurement"
STATE_CLASS_TOTAL_INCREASING = "total_increasing"


class mqtt(Plugin):
    def __init__(self):
        super().__init__()
        self.name = "mqtt"
        self.description = "Write output to internal mqtt facility of Home Assistant"
        tz = self.config.get("default", "timezone", fallback="Europe/Amsterdam")
        self.timezone = pytz.timezone(tz)

        self.mqtt_client_name = (
            self.config.get(
                "output.mqtt", "client_name_prefix", fallback="ha-mqtt-omniklogger"
            )
            + "_"
            + uuid.uuid4().hex
        )
        self.mqtt_host = self.config.get("output.mqtt", "host", fallback="localhost")
        self.mqtt_port = int(self.config.get("output.mqtt", "port", fallback="1833"))
        self.mqtt_retain = self.config.getboolean(
            "output.mqtt", "retain", fallback=False
        )
        self.mqtt_config_published = {}
        if not self.config.has_option(
            "output.mqtt", "username"
        ) or not self.config.has_option("output.mqtt", "password"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "Please specify MQTT username and password in the configuration",
            )
            self.mqtt_username = None
            self.mqtt_password = None
        else:
            self.mqtt_username = self.config.get("output.mqtt", "username")
            self.mqtt_password = self.config.get("output.mqtt", "password")

        # mqtt setup
        self.mqtt_client = mqttclient.Client(self.mqtt_client_name)
        self.mqtt_client.on_connect = self._mqtt_on_connect  # bind call back function
        self.mqtt_client.on_disconnect = (
            self._mqtt_on_disconnect
        )  # bind call back function
        self.mqtt_client.logger = self.logger
        self.mqtt_client.hass_api = self.hass_api
        # self.mqtt_client.on_message=mqtt_on_message (not used)
        self.mqtt_client.username_pw_set(self.mqtt_username, self.mqtt_password)
        if self.config.has_option("output.mqtt", "discovery_prefix"):
            self.discovery_prefix = self.config.get("output.mqtt", "discovery_prefix")
        else:
            self.discovery_prefix = "homeassistant"
        # Init topics dict
        self.topics = {}
        # Init config dict
        self.config_pl = {}
        # Make instance to run exclusively
        self.access = threading.Condition(threading.Lock())
        self._connected = self.mqtt_connect()

    def mqtt_connect(self) -> bool:
        try:
            # connect
            self.mqtt_client.connect(self.mqtt_host, self.mqtt_port)
            # start processing messages
            self.mqtt_client.loop_start()
            return True
        except OSError as ex:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Unable to connect to MQTT server, is your configuration correct? MQTT was not initialized correctly. Error: {ex}",
            )
            return False

    def _init_config(self, msg):
        # Check if init is needed
        asset_classes = set()
        for field in msg:
            if field in self.config.data_field_config:
                asset_class = self.config.data_field_config[field]["asset"]
                if asset_class not in asset_classes:
                    asset_classes.add(asset_class)

        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        # field: name, dev_cla, ic, unit, measurement, filter, asset

        # Assemble topics
        self.topics[msg["plant_id"]] = self._topics(msg, asset_classes)

        # Assemble config
        self.config_pl[msg["plant_id"]] = self._config_payload(
            msg, self.topics[msg["plant_id"]], asset_classes
        )

        # Flag init as done
        # self.mqtt_field_name_override_init[msg['plant_id']] = True
        return asset_classes

    def _mqtt_on_connect(self, client, userdata, flags, rc=0, properties=None):
        if rc == 0:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "MQTT connected")
            # subscribe listening (not used)

    def _mqtt_on_disconnect(self, client, userdata, flags, rc=0):
        if rc == 0:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "MQTT disconnected")

    def _topics(self, msg, asset_classes):
        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        # field: name, dev_cla, ic, unit, filter
        topic_suffix = f"_{msg['plant_id']}" if not msg["plant_id"] == "0" else ""
        topics = {}
        for asset_class in asset_classes:
            topics[asset_class] = {}
            topics[asset_class][
                "main"
            ] = f"{self.discovery_prefix}/sensor/{self.config.attributes['devicename'][asset_class]}{topic_suffix}"
            topics[asset_class]["state"] = f"{topics[asset_class]['main']}/state"
            topics[asset_class]["attr"] = f"{topics[asset_class]['main']}/attr"
            topics[asset_class]["config"] = {}
            for field in self.config.data_field_config:
                if field in msg:
                    topics[asset_class]["config"][
                        field
                    ] = f"{topics[asset_class]['main']}/{field}/config"
        return topics

    def _device_payload(self, msg, asset_classes):
        device_pl = {}
        for asset_class in asset_classes:
            device_pl[asset_class] = {}
            # Set device_name
            device_name = self.config.attributes["devicename"][asset_class]
            # Determine appendixes
            name_appendix = ""
            identifier = self.config.attributes["identifier"][asset_class]
            if identifier in msg and self.config.getboolean(
                "output.mqtt", "append_plant_id", False
            ):
                name_appendix = f" {msg[identifier]}"
            id_prefix = ""
            if identifier in msg:
                id_prefix = f"{msg[identifier]}_"
            # Device payload
            device_pl[asset_class] = {
                "identifiers": [f"{id_prefix}{asset_class}"],
                "name": f"{device_name}{name_appendix}",
                "mdl": self.config.attributes["model"][asset_class],
                "mf": self.config.attributes["mf"][asset_class],
            }
        return device_pl

    def _config_payload(self, msg, topics, asset_classes):
        # Get device payload
        device_pl = self._device_payload(msg, asset_classes)

        # Fill config_pl dict
        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        config_pl = {}

        # Generate MQTT config topics
        for field in self.config.data_field_config:
            # Only generate config for available data
            if field not in msg:
                # skip publishing non exitent values
                continue
            asset_class = self.config.data_field_config[field]["asset"]
            if not asset_class:
                # skip publication
                continue
            # field: name, dev_cla, ic, unit, measurent, filter, asset
            if self.config.has_option("output.mqtt", f"{field}_name"):
                fieldname = self.config.get("output.mqtt", f"{field}_name")
            else:
                fieldname = self.config.data_field_config[field]["name"]
            identifier = self.config.attributes["identifier"][asset_class]
            config_pl[field] = {
                "~": f"{topics[asset_class]['main']}",
                "uniq_id": f"{msg[identifier]}_{field}",
                "name": f"{fieldname}",
                "stat_t": "~/state",
                "json_attr_t": "~/attr",
                "val_tpl": f"{{{{(value_json.{field}{self.config.data_field_config[field]['filter']})}}}}",
                "dev": device_pl[asset_class],
            }

            # Set device and state class if configured
            if self.config.data_field_config[field]["dev_cla"]:
                config_pl[field]["dev_cla"] = self.config.data_field_config[field][
                    "dev_cla"
                ]
            if (
                self.config.data_field_config[field]["dev_cla"]
                in MEASUREMENT_TOTAL_INCREASING_CLASSES
            ):
                state_class = STATE_CLASS_TOTAL_INCREASING
            elif self.config.data_field_config[field]["dev_cla"] in MEASUREMENT_CLASSES:
                state_class = STATE_CLASS_MEASUREMENT
            else:
                state_class = None

            # set the state class for measurements
            if state_class:
                config_pl[field]["stat_cla"] = state_class
            # Set icon if configured
            if self.config.data_field_config[field]["ic"]:
                config_pl[field][
                    "ic"
                ] = f"mdi:{self.config.data_field_config[field]['ic']}"
            # Set unit of measurement if configured
            if self.config.data_field_config[field]["unit"]:
                config_pl[field]["unit_of_meas"] = self.config.data_field_config[field][
                    "unit"
                ]

        return config_pl

    def _value_payload(self, msg):
        value_pl = {}
        # Generate MQTT config topics
        for field in self.config.data_field_config:
            # field: name, dev_cla, ic, unit, filter
            # Only generate payload for available data
            asset_class = self.config.data_field_config[field]["asset"]
            if not asset_class:
                # skip publication
                continue
            if asset_class not in value_pl:
                value_pl[asset_class] = {}
            if field in msg:
                value_pl[asset_class][field] = self.jsonval(msg[field])

        return value_pl

    def _attribute_payload(self, msg, asset_classes):
        attr_pl = {}
        for asset_class in asset_classes:
            apl = self._attribute_payload_asset_class(msg, asset_class)
            if apl:
                attr_pl[asset_class] = apl
        return attr_pl

    def _attribute_payload_asset_class(self, msg, asset_class):
        attr_pl_class = {}
        for attr in self.config.attributes["asset"][asset_class]:
            attr_pl = self._encode_attribute(msg, attr)
            if attr_pl:
                attr_pl_class[attr] = self._encode_attribute(msg, attr)
        return attr_pl_class

    def _encode_attribute(self, msg, attr):
        if attr not in msg:
            return None
        if attr in self.config.data_field_config:
            return (
                time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(msg[attr]))
                if self.config.data_field_config[attr]["dev_cla"] == "timestamp"
                else self.jsonval(msg[attr])
            )
        else:
            return self.jsonval(msg[attr])

    def _publish_config(self, msg):
        new_asset_classes = set()
        if msg["plant_id"] not in self.mqtt_config_published:
            # init mqtt_config_published flag cache
            self.mqtt_config_published[msg["plant_id"]] = set()

        for entity in self.config_pl[msg["plant_id"]]:
            asset_class = self.config.data_field_config[entity]["asset"]
            if (
                asset_class in self.mqtt_config_published[msg["plant_id"]]
                and self.mqtt_retain
            ):
                continue
            else:
                self._publish_config_entity(
                    self.topics[msg["plant_id"]][asset_class],
                    self.config_pl[msg["plant_id"]],
                    entity,
                )
                new_asset_classes.add(asset_class)
        for item in new_asset_classes:
            self.mqtt_config_published[msg["plant_id"]].add(item)

    def _publish_config_entity(self, topics, config_pl, entity):
        try:
            # publish config
            # asset_class = self.config.data_field_config[entity]['asset']
            if self.mqtt_client.publish(
                topics["config"][entity],
                json.dumps(config_pl[entity]),
                retain=self.mqtt_retain,
            ):
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    f"Publishing config {json.dumps(config_pl[entity])} "
                    f"to {topics['config'][entity]} successful.",
                )
            else:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "WARNING",
                    f"Publishing config {json.dumps(config_pl[entity])} "
                    f"to {topics['config'][entity]} failed!",
                )
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Unhandled error publishing config for entity {entity}: {e}",
            )

    def _publish_attributes(self, msg, asset_classes):

        attr_pl = self._attribute_payload(msg, asset_classes)
        for asset_class in asset_classes:
            try:
                # publish attributes
                if self.mqtt_client.publish(
                    self.topics[msg["plant_id"]][asset_class]["attr"],
                    json.dumps(attr_pl[asset_class]),
                    retain=self.mqtt_retain,
                ):
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "DEBUG",
                        f"Publishing attributes {json.dumps(attr_pl[asset_class])} to "
                        f"{self.topics[msg['plant_id']][asset_class]['attr']} successful.",
                    )
                else:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "WARNING",
                        f"Publishing attributes {json.dumps(attr_pl[asset_class])} to "
                        f"{self.topics[msg['plant_id']][asset_class]['attr']} failed!",
                    )
            except Exception as e:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    f"Unhandled error publishing attributes: {e}",
                )

    def _publish_state(self, topics, value_pl, asset_classes):
        for asset_class in asset_classes:
            try:
                # publish state
                if self.mqtt_client.publish(
                    topics[asset_class]["state"],
                    json.dumps(value_pl[asset_class]),
                    retain=self.mqtt_retain,
                ):
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "DEBUG",
                        f"Publishing state {json.dumps(value_pl[asset_class])} to "
                        f"{topics[asset_class]['state']} successful.",
                    )
                else:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "WARNING",
                        f"Publishing state {json.dumps(value_pl[asset_class])} to "
                        f"{topics[asset_class]['state']} failed!",
                    )
            except Exception as e:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    f"Unhandled error publishing states: {e}",
                )

    def process(self, **args):
        """
        Send data to over mqtt (compliant with Home Assistant MQTT discovery standard
        See: https://www.home-assistant.io/docs/mqtt/discovery/)
        """

        # Check if we have a valid loop
        if not self._connected:
            self._connected = self.mqtt_connect()
            if not self._connected:
                return

        self.access.acquire()

        # Get argument
        msg = args["msg"]

        # Assemble config
        asset_classes = self._init_config(msg)

        # Publish config
        self._publish_config(msg)

        # publish attributes
        self._publish_attributes(msg, asset_classes)

        # publish state
        value_pl = self._value_payload(msg)
        self._publish_state(self.topics[msg["plant_id"]], value_pl, asset_classes)

        self.access.release()



# Credits to: https://github.com/Woutrrr
# adjusted for python 3.5+

import struct  # Converting bytes to numbers
import datetime
import time


class InverterMsg:

    "Class for Inverter message"
    rawmsg = ""

    def __init__(self, msg, offset=0):
        self.rawmsg = msg
        self.offset = offset
        # Set a timestamp
        self.last_update = time.time()

    def __getString(self, begin, end):
        return str(self.rawmsg[begin:end], encoding="UTF-8")

    def __getShort(self, begin, devider=10):
        num = struct.unpack("!H", self.rawmsg[begin : begin + 2])[0]
        if num == 65535:
            return -1
        else:
            return float(num) / devider

    def __getLong(self, begin, devider=10):
        return float(struct.unpack("!I", self.rawmsg[begin : begin + 4])[0]) / devider

    def getID(self):
        return self.__getString(15, 31)

    def getTemp(self):
        return self.__getShort(31)

    def getPower(self):
        return int(self.__getShort(59, 1))

    def getPVPower(self):
        totalpower = 0
        strings = [1, 2, 3]
        for i in strings:
            if self.getPPV(i) > 0:
                totalpower += self.getPPV(i)
        return totalpower

    def getETotal(self):
        return self.__getLong(71)

    def getVPV(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 33 + (i - 1) * 2
        return self.__getShort(num)

    def getIPV(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 39 + (i - 1) * 2
        return self.__getShort(num)

    def getPPV(self, i=1):
        # If current is 0 return 0
        if not self.getIPV(i):
            return 0
        # If current is -1 (invalid) return -1 (invalid)
        if self.getIPV(i) < 0:
            return -1
        else:
            return int(self.getVPV(i) * self.getIPV(i))

    def getIAC(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 45 + (i - 1) * 2
        return self.__getShort(num)

    def getVAC(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 51 + (i - 1) * 2
        return self.__getShort(num)

    def getFAC(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 57 + (i - 1) * 4
        return self.__getShort(num, 100)

    def getPAC(self, i=1):
        if i not in range(1, 4):
            i = 1
        num = 59 + (i - 1) * 4
        return int(self.__getShort(num, 1))  # Don't divide

    def getEToday(self):
        return self.__getShort(69, 100)  # Devide by 100

    def getHTotal(self):
        return int(self.__getLong(75, 1))  # Don't divide

    def getIACalt(self, i=1):
        # Calculate current from power and voltage
        pac = self.getPAC(i)
        vac = self.getVAC(i)
        return round(pac / vac, 2) if pac >= 0 and vac > 0 else -1

    def setIfValid(self, key, data, value):
        if value >= 0:
            data[key] = value

    def _maxoflist(self, list):
        max = -1
        for item in list:
            if item > max:
                max = item
        return max

    def FetchDataDict(self, data):
        if data:
            if not isinstance(data, dict):
                # Object is not a dict
                raise Exception("data object is not a dict")
        else:
            data = {}
        # Fill the dict
        # Get timestamp string and epoch
        data["last_update_time"] = datetime.datetime.utcfromtimestamp(
            self.last_update
        ).strftime("%Y-%m-%dT%H:%M:%SZ")
        data["last_update"] = self.last_update
        # Set the serial number
        data["inverter"] = self.getID()
        data["current_power"] = self.getPower()
        data["today_energy"] = self.getEToday()
        data["total_energy"] = self.getETotal()
        data["inverter_temperature"] = self.getTemp()
        self.setIfValid("current_ac1", data, self.getIACalt(1))
        self.setIfValid("current_ac2", data, self.getIACalt(2))
        self.setIfValid("current_ac3", data, self.getIACalt(3))
        self.setIfValid("voltage_ac1", data, self.getVAC(1))
        self.setIfValid("voltage_ac2", data, self.getVAC(2))
        self.setIfValid("voltage_ac3", data, self.getVAC(3))
        self.setIfValid("frequency_ac1", data, self.getFAC(1))
        self.setIfValid("frequency_ac2", data, self.getFAC(2))
        self.setIfValid("frequency_ac3", data, self.getFAC(3))
        self.setIfValid("power_ac1", data, self.getPAC(1))
        self.setIfValid("power_ac2", data, self.getPAC(2))
        self.setIfValid("power_ac3", data, self.getPAC(3))
        self.setIfValid("voltage_pv1", data, self.getVPV(1))
        self.setIfValid("voltage_pv2", data, self.getVPV(2))
        self.setIfValid("voltage_pv3", data, self.getVPV(3))
        self.setIfValid("current_pv1", data, self.getIPV(1))
        self.setIfValid("current_pv2", data, self.getIPV(2))
        self.setIfValid("current_pv3", data, self.getIPV(3))
        self.setIfValid("power_pv1", data, self.getPPV(1))
        self.setIfValid("power_pv2", data, self.getPPV(2))
        self.setIfValid("power_pv3", data, self.getPPV(3))
        self.setIfValid("operation_hours", data, self.getHTotal())
        self.setIfValid("current_power_pv", data, self.getPVPower())
        self.setIfValid(
            "voltage_ac_max",
            data,
            self._maxoflist([self.getVAC(1), self.getVAC(2), self.getVAC(3)]),
        )


def request_string(ser):
    """
    The request string is build from several parts. The first part is a
    fixed 4 char string; the second part is the reversed hex notation of
    the Wi-Fi logger s/n twice; then again a fixed string of two chars; a checksum of
    the double s/n with an offset; and finally a fixed ending char.
    this code requires python 3.5 or later!
    """
    responseString = b"\x68\x02\x40\x30"

    doublehex = hex(ser)[2:] * 2
    hexlist = [
        bytes.fromhex(doublehex[i : i + 2])
        for i in reversed(range(0, len(doublehex), 2))
    ]

    cs_count = 115 + sum([ord(c) for c in hexlist])
    cs = bytes.fromhex(hex(cs_count)[-2:])
    responseString += b"".join(hexlist) + b"".join([b"\x01\x00", cs, b"\x16"])
    return responseString
